package fullenv

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"path"
	"path/filepath"
	"reflect"
	"regexp"
	"sort"
	"strings"
	"syscall"
	"testing"
	"time"

	"github.com/tgulacsi/go-locking"
	"gopkg.in/yaml.v2"
	// Embed QB dialects we need
	_ "github.com/slicebit/qb/dialects/postgres"
	"github.com/termie/go-shutil"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"

	_ "github.com/lib/pq" // full env tests run on postgres
)

// DefaultBinaryPaths defines the default paths for the various binaries
var DefaultBinaryPaths = make(map[string]string)

// RawSettings holds the raw configuration of a xbus program (xbus, xbus-client...)
type RawSettings yaml.MapSlice

func mapSliceSet(s *RawSettings, key string, value interface{}) {
	for i, item := range *s {
		sKey, ok := item.Key.(string)
		if !ok {
			panic("key is not a string")
		}
		if sKey == key {
			item.Value = value
			(*s)[i] = item
			return
		}
	}
	if !strings.Contains(key, ".") {
		*s = append(*s, yaml.MapItem{Key: key, Value: value})
		return
	}
	splittedKey := strings.SplitN(key, ".", 2)
	root := splittedKey[0]
	path := splittedKey[1]
	for i, item := range *s {
		sKey, _ := item.Key.(string)
		if sKey == root {
			submap, ok := item.Value.(RawSettings)
			if !ok {
				panic(fmt.Sprintf("%s is not a rawsettings: %v", root, reflect.TypeOf(item.Value)))
			}
			mapSliceSet(&submap, path, value)
			item.Value = submap
			(*s)[i] = item
			return
		}
	}
	v := RawSettings{}
	mapSliceSet(&v, path, value)
	*s = append(*s, yaml.MapItem{Key: root, Value: v})
}

func mapSliceGet(s *RawSettings, key string) interface{} {
	for _, item := range *s {
		sKey, ok := item.Key.(string)
		if !ok {
			panic("key is not a string")
		}
		if sKey == key {
			return item.Value
		}
	}
	if !strings.Contains(key, ".") {
		return nil
	}
	splittedKey := strings.SplitN(key, ".", 2)
	root := splittedKey[0]
	path := splittedKey[1]
	rootV := mapSliceGet(s, root)
	if rootV == nil {
		return nil
	}
	msRoot, ok := rootV.(RawSettings)
	if !ok {
		return nil
	}
	return mapSliceGet(&msRoot, path)
}

// FileCopyOption holds options to copy files to the test env
type FileCopyOption struct {
	Source string
	Target string
}

// FileOption holds options to setup files to the test env
type FileOption struct {
	Copy *FileCopyOption
}

// Options holds all the options necessary to setup a FullEnv
type Options struct {
	Xbusd   RawSettings
	Clients []RawSettings
	HTTP    []RawSettings
	Graphs  map[string]RawSettings
	Dsn     string

	// If true, no attempt to prevent concurrent access is done.
	NoDBLock bool

	TargetDir string

	ConfigFilePath string
	Files          []FileOption

	BinaryPaths map[string]string

	Logger xbus.Logger
}

// OptionsFromFile initialize a Options from a yaml configuration file
func OptionsFromFile(configfile string) (Options, error) {
	buf, err := ioutil.ReadFile(configfile)
	if err != nil {
		return Options{}, err
	}

	var options Options
	if err := yaml.Unmarshal(buf, &options); err != nil {
		return options, err
	}
	options.ConfigFilePath = configfile

	return options, nil
}

// NewFullEnvForTest returns a FullEnv that sends logs to a *testing.T
func NewFullEnvForTest(t *testing.T, basename, configfile string) (*FullEnv, *testutils.TestLogger) {
	options, err := OptionsFromFile(configfile)
	if err != nil {
		t.Fatal(err)
	}
	logger := testutils.GetTestLogger(t)
	options.Logger = logger

	env := NewFullEnv(basename, options)
	env.t = t
	return env, logger
}

// NewFullEnv initializes a FullEnv
func NewFullEnv(basename string, options Options) *FullEnv {
	if options.BinaryPaths == nil {
		options.BinaryPaths = make(map[string]string)
	}
	for name, path := range DefaultBinaryPaths {
		if _, ok := options.BinaryPaths[name]; !ok {
			options.BinaryPaths[name] = path
		}
	}
	return &FullEnv{
		basename: basename,
		options:  options,
		log:      options.Logger,

		programs: make(map[string]program),
	}
}

// ProgramSettings defines how and when to start a program
type ProgramSettings struct {
	Autostart bool
	Priority  int
}

type program interface {
	Settings() ProgramSettings
	Running() bool
	Setup(xbusctl CmdRunner) error
	Start() error
	Stop() error
}

// FullEnv pilots xbusd and clients
type FullEnv struct {
	basename string
	options  Options

	t       *testing.T
	log     xbus.Logger
	dirname string
	dsn     string
	dblock  *locking.PortLock

	actorServices []string

	xbusd   *exec.Cmd
	XbusCtl CmdRunner
	Xbusd   CmdRunner

	programs        map[string]program
	runningPrograms []program

	Emitter  *Emitter
	Consumer *Consumer
}

// RegisterActorService registers an actor service for this env only
// Allow to inject actors that are controlled by the test
func (fe *FullEnv) RegisterActorService(name string, init xbus.NewActorServiceFunc) {
	fe.actorServices = append(fe.actorServices, name)
	xbus.RegisterActorService(name, init)
}

// BinPath returns the path of a binary.
func (fe *FullEnv) BinPath(name string) string {
	if path, ok := fe.options.BinaryPaths[name]; ok {
		return path
	}
	// if GOPATH is present, use it
	if p := os.Getenv("GOPATH"); p != "" {
		return path.Join(p, "bin", name)
	}

	// if not, classic PATH lookup
	return name
}

func (fe *FullEnv) getDsn() (string, error) {

	if fe.options.Dsn != "" {
		return fe.options.Dsn, nil
	}

	var dsn string
	var ok bool

	dsnConf := mapSliceGet(&fe.options.Xbusd, "database.dsn")

	if dsnConf == nil {
		if dsn, ok = os.LookupEnv("XBUS_TEST_POSTGRES_DSN"); !ok {
			return "", errors.New(
				"Need a XBUS_TEST_POSTGRES_DSN environment variable")
		}
		return dsn, nil
	}

	if dsn, ok = dsnConf.(string); !ok {
		return "", errors.New("xbusd dsn configuration type should be string")
	}

	return dsn, nil
}

// SetupDB provides an empty database in the database.dsn xbusd setting
func (fe *FullEnv) SetupDB() error {

	var dsn string
	var err error

	if dsn, err = fe.getDsn(); err != nil {
		return err
	}

	if !fe.options.NoDBLock {
		fe.dblock = locking.NewPortLock(1764)
		if err := fe.dblock.Lock(); err != nil {
			fe.log.Errorf("Cannot get lock: %s", err)
			return err
		}
	}

	_, db, err := model.NewDB("postgres", dsn, 2, fe.log, 0)
	if err != nil {
		return err
	}

	// Cleanup the db
	tables := db.Metadata.GetQbMetadata().Tables()
	for i := len(tables) - 1; i >= 0; i-- {
		table := tables[i]
		drop := table.Drop(db.Engine.Dialect())
		if _, err := db.Engine.DB().Exec(drop); err != nil {
			fe.log.Noticef("%s", err)
		}

		if _, err := db.Engine.DB().DB.Exec("DROP TABLE schema_migrations"); err != nil {
			fe.log.Noticef("%s", err)
		}
	}

	db.Close()

	fe.dsn = dsn

	mapSliceSet(&fe.options.Xbusd, "database.dsn", dsn)

	return nil
}

func (fe *FullEnv) runXbusd(args ...string) error {
	xbusdPath := fe.BinPath("xbusd")

	xbusdCfgFilename := path.Join(fe.dirname, "xbusd.yaml")

	args = append([]string{
		"--config", xbusdCfgFilename,
	}, args...)

	cmd := exec.Command(xbusdPath, args...)
	cmd.Dir = fe.dirname

	b, err := cmd.CombinedOutput()
	if err != nil {
		fe.log.Errorf("%s", b)
		return fmt.Errorf("Could not run xbusd : %s", err)
	}

	return nil
}

// SetupServer runs the commands for setting up an operational xbusd
func (fe *FullEnv) SetupServer() error {
	if err := fe.Xbusd.Run("init", "--debug"); err != nil {
		return err
	}

	if err := fe.Xbusd.Run("key", "generate"); err != nil {
		return err
	}

	if err := fe.Xbusd.Run("cert", "generate", "server-ca"); err != nil {
		return err
	}

	if err := fe.Xbusd.Run("cert", "generate", "client-ca"); err != nil {
		return err
	}

	return fe.StartXbusd()
}

// SetupCtl runs the commands for setting up a operational XbusCtl
func (fe *FullEnv) SetupCtl() (err error) {
	defer catcher(&err)

	fe.XbusCtl.MustRun("key", "generate")
	fe.XbusCtl.MustRun("register")

	must(fe.Xbusd.Run("account", "accept", "ctl"))

	fe.XbusCtl.MustRun("register")

	return nil
}

func (fe *FullEnv) setupGraphs() (err error) {
	defer catcher(&err)

	graphDir := path.Join(fe.dirname, "graphs")
	must(os.Mkdir(graphDir, 0700))

	for name, graphDef := range fe.options.Graphs {
		graph, err := yaml.Marshal(graphDef)
		must(err)

		version := "0"
		status := "active"

		if strings.Contains(name, "#") {
			splitted := strings.SplitN(name, "#", 2)
			name = splitted[0]
			status = splitted[1]
		}
		if strings.Contains(name, "-") {
			splitted := strings.SplitN(name, "-", 2)
			name = splitted[0]
			version = splitted[1]
		}

		graphfile := path.Join(graphDir, name+".yaml")
		must(ioutil.WriteFile(graphfile, graph, 0600))
		fe.XbusCtl.MustRun("graph", "save", name, version, graphfile)
		switch status {
		case "active":
			fe.XbusCtl.MustRun("graph", "status", name, version, "--active")
		case "inactive":
			fe.XbusCtl.MustRun("graph", "status", name, version, "--inactive")
		default:
		}
	}
	return nil
}

func (fe *FullEnv) writeClientConf(
	config RawSettings,
) (
	settings clientSettings, err error,
) {
	settings.autostart = true
	var removeEntries []int
	for i, item := range config {
		if s, ok := item.Key.(string); ok {
			switch s {
			case "account-name":
				if s, ok := item.Value.(string); ok {
					settings.name = s
				}
			case "inprocess":
				if s, ok := item.Value.(bool); ok {
					settings.inProcess = s
					removeEntries = append(removeEntries, i)
				}
			case "custom-xbus-client":
				if s, ok := item.Value.(string); ok {
					settings.customXbusClient = s
					removeEntries = append(removeEntries, i)
				}
			case "autostart":
				if s, ok := item.Value.(bool); ok {
					settings.autostart = s
					removeEntries = append(removeEntries, i)
				}
			case "apikey":
				if s, ok := item.Value.(string); ok {
					settings.apikey = s
					removeEntries = append(removeEntries, i)
				}
			}
		}
	}

	for i := len(removeEntries) - 1; i >= 0; i-- {
		entry := removeEntries[i]
		config = append(config[:entry], config[entry+1:]...)
	}

	var (
		// TODO forbid names with path separators
		configDir  = path.Join(fe.dirname, "client-"+settings.name)
		configFile = "xbus-client.yaml"
	)
	if err = os.MkdirAll(configDir, 0700); err != nil {
		return
	}
	var b []byte
	b, err = yaml.Marshal(config)
	if err != nil {
		return
	}
	if err = ioutil.WriteFile(path.Join(configDir, configFile), b, 0600); err != nil {
		return
	}

	return
}

func (fe *FullEnv) writeHTTPConf(
	config RawSettings,
) (
	settings httpSettings, err error,
) {
	settings.autostart = true
	for _, item := range config {
		if s, ok := item.Key.(string); ok {
			if s == "account-name" {
				if s, ok := item.Value.(string); ok {
					settings.name = s
				}
			}
			if s == "autostart" {
				if s, ok := item.Value.(bool); ok {
					settings.autostart = s
				}
			}
		}
	}

	var (
		// TODO forbid names with path separators
		configDir  = path.Join(fe.dirname, "http-"+settings.name)
		configFile = "xbus-http.yaml"
	)
	if err = os.MkdirAll(configDir, 0700); err != nil {
		return
	}
	var b []byte
	b, err = yaml.Marshal(config)
	if err != nil {
		return
	}
	if err = ioutil.WriteFile(path.Join(configDir, configFile), b, 0600); err != nil {
		return
	}

	return
}

func (fe *FullEnv) setupClient(settings clientSettings) error {
	var (
		name = settings.name
		// TODO forbid names with path separators
		configDir  = path.Join(fe.dirname, "client-"+name)
		configFile = "xbus-client.yaml"
	)

	client := clientRunner{
		dirname:       configDir,
		configFile:    configFile,
		xbusClientBin: fe.BinPath("xbus-client"),
		logger:        fe.options.Logger,
		settings:      settings,
		cmd:           nil,
		inClient:      nil,
		running:       false,
	}

	fe.programs[name] = &client

	return client.Setup(fe.XbusCtl)
}

func (fe *FullEnv) setupHTTP(settings httpSettings) error {
	var (
		name = settings.name
		// TODO forbid names with path separators
		configDir  = path.Join(fe.dirname, "http-"+name)
		configFile = "xbus-http.yaml"
	)

	http := httpRunner{
		dirname:     configDir,
		configFile:  configFile,
		xbusHTTPBin: fe.BinPath("xbus-http"),
		logger:      fe.options.Logger,
		settings:    settings,
		cmd:         nil,
		running:     false,
	}

	fe.programs[name] = &http

	return http.Setup(fe.XbusCtl)
}

func (fe *FullEnv) setupClients() error {
	for _, config := range fe.options.Clients {
		settings, err := fe.writeClientConf(config)
		if err != nil {
			return err
		}
		if err := fe.setupClient(settings); err != nil {
			return err
		}
	}
	for _, config := range fe.options.HTTP {
		settings, err := fe.writeHTTPConf(config)
		if err != nil {
			return err
		}
		if err := fe.setupHTTP(settings); err != nil {
			return err
		}
	}
	return nil
}

// WorkDir returns the directory where all files are written
func (fe *FullEnv) WorkDir() string {
	return fe.dirname
}

// CreateEnv creates the test directories, configuration files,
// setup the XbusCtl and Xbusd command runners.
// It does not run any command (see Init)
func (fe *FullEnv) CreateEnv() error {
	var dirname = fe.options.TargetDir

	if dirname != "" {
		var err error
		dirname, err = filepath.Abs(dirname)
		if err != nil {
			return err
		}
	} else {
		var err error

		if dirname, err = ioutil.TempDir(
			"", "xbus-fullenv-tests-"+fe.basename+"--",
		); err != nil {
			return err
		}
	}

	if err := os.MkdirAll(path.Join(dirname, "certs", "xbusd"), 0700); err != nil {
		return err
	}

	fe.log.Noticef("Fullenv working dir created: %s", dirname)

	fe.dirname = dirname

	// xbusd conf
	conf, err := yaml.Marshal(fe.options.Xbusd)
	if err != nil {
		panic(err)
	}
	xbusdCfgFilename := path.Join(fe.dirname, "xbusd.yaml")
	if err := ioutil.WriteFile(xbusdCfgFilename, conf, 0600); err != nil {
		return err
	}

	fe.Xbusd = NewCmdRunner(
		xbus.NewPrefixedLogger("-- xbusd ", fe.log),
		fe.dirname, fe.BinPath("xbusd"),
		"--config", xbusdCfgFilename,
		"--cert-dir", path.Join(fe.dirname, "certs", "xbusd"))

	// xbusCtl conf
	dirname = path.Join(fe.dirname, "xbusctl")
	if err := os.Mkdir(dirname, 0700); err != nil {
		return err
	}

	xbusdCtlFilename := path.Join(dirname, "xbusctl.yaml")
	binpath := fe.BinPath("xbusctl")

	if err := ioutil.WriteFile(xbusdCtlFilename, []byte("account-name: ctl"), 0600); err != nil {
		return err
	}

	fe.XbusCtl = NewCmdRunner(
		xbus.NewPrefixedLogger("-- xbusctl --", fe.log),
		dirname, binpath, "--config", xbusdCtlFilename)

	return nil
}

// SetupFiles copy all the files specified in the 'files' section
func (fe *FullEnv) SetupFiles() error {
	configDir := "."
	if fe.options.ConfigFilePath != "" {
		configDir = path.Dir(fe.options.ConfigFilePath)
	}
	for _, fileOpt := range fe.options.Files {
		if fileOpt.Copy != nil && fileOpt.Copy.Source != "" {
			source := path.Join(configDir, fileOpt.Copy.Source)
			target := fileOpt.Copy.Target
			if target == "" {
				target = path.Base(source)
			}
			target = path.Join(fe.dirname, fileOpt.Copy.Target)
			targetDir := path.Dir(target)
			if _, err := os.Stat(targetDir); os.IsNotExist(err) {
				if err := os.MkdirAll(targetDir, os.ModePerm); err != nil {
					return err
				}
			}
			if err := shutil.CopyFile(source, target, true); err != nil {
				return err
			}
			fe.log.Noticef("Copied %s to %s", source, target)
		}
	}
	return nil
}

// SetupInternalActors initializes the fullenv actors
func (fe *FullEnv) SetupInternalActors() {
	fe.Emitter = NewEmitter(fe.log)
	fe.Consumer = NewConsumer(fe.log)
	fe.RegisterActorService("fullenv.emitter", fe.Emitter.Service)
	fe.RegisterActorService("fullenv.consumer", fe.Consumer.Service)
}

// Init setup the server, the clients and load the graphs
// Once done only the server is running
func (fe *FullEnv) Init() error {
	var success bool
	defer func(s *bool) {
		if !*s {
			fe.Close()
		}
	}(&success)

	if err := fe.SetupDB(); err != nil {
		return err
	}

	if err := fe.CreateEnv(); err != nil {
		return err
	}

	if err := fe.SetupFiles(); err != nil {
		return err
	}

	fe.SetupInternalActors()

	if err := fe.SetupServer(); err != nil {
		return err
	}
	if err := fe.SetupCtl(); err != nil {
		return err
	}
	if err := fe.setupClients(); err != nil {
		return err
	}
	if err := fe.setupGraphs(); err != nil {
		return err
	}
	success = true
	return nil
}

// StartClient starts a client
func (fe *FullEnv) StartClient(name string) error {
	if c, ok := fe.programs[name]; ok {
		return c.Start()
	}
	return fmt.Errorf("No such client: %s", name)
}

// StopClient stops a client
func (fe *FullEnv) StopClient(name string) error {
	if c, ok := fe.programs[name]; ok {
		return c.Stop()
	}
	return fmt.Errorf("No such client: %s", name)
}

// ClientConfig returns the configuration file of a client
func (fe *FullEnv) ClientConfig(name string) (string, error) {
	if c := fe.getClientRunner(name); c != nil {
		return c.configPath(), nil
	}
	return "", fmt.Errorf("No such client: %s", name)
}

// GetClient returns the xbus.Client instance corresponding to the inprocess
// client
func (fe *FullEnv) GetClient(name string) *xbus.Client {
	if c := fe.getClientRunner(name); c != nil {
		return c.inClient
	}
	return nil
}

func (fe *FullEnv) getClientRunner(name string) *clientRunner {
	p, _ := fe.programs[name]
	if p != nil {
		if c, ok := p.(*clientRunner); ok {
			return c
		}
	}
	return nil
}

// Startup starts the programs
func (fe *FullEnv) Startup() error {
	var (
		programs   = make(map[int][]program)
		priorities []int
	)
	for _, p := range fe.programs {
		programs[p.Settings().Priority] = append(
			programs[p.Settings().Priority], p)
	}
	for priority := range programs {
		priorities = append(priorities, priority)
	}
	sort.Ints(priorities)

	for _, priority := range priorities {
		for _, p := range programs[priority] {
			if !p.Settings().Autostart {
				continue
			}
			if err := p.Start(); err != nil {
				return err
			}
			fe.runningPrograms = append(fe.runningPrograms, p)
		}
	}
	return nil
}

// Shutdown stops the programs
func (fe *FullEnv) Shutdown() error {
	for i := len(fe.runningPrograms) - 1; i >= 0; i-- {
		p := fe.runningPrograms[i]
		p.Stop()
		fe.runningPrograms = fe.runningPrograms[:len(fe.runningPrograms)-1]
	}
	return nil
}

func captureCmdOutputs(name string, cmd *exec.Cmd, log xbus.Logger) error {
	stdout, err := cmd.StdoutPipe()
	if err != nil {
		return err
	}

	stderr, err := cmd.StderrPipe()
	if err != nil {
		return err
	}

	go func() {
		scanner := bufio.NewScanner(stdout)
		for scanner.Scan() {
			log.Noticef("-- " + name + ":out -- " + scanner.Text())
		}
	}()

	go func() {
		scanner := bufio.NewScanner(stderr)
		for scanner.Scan() {
			log.Noticef("-- " + name + ":err -- " + scanner.Text())
		}
	}()
	return nil
}

// StartXbusd starts the xbusd server
func (fe *FullEnv) StartXbusd() error {
	xbusdCfgFilename := path.Join(fe.dirname, "xbusd.yaml")
	xbusd := exec.Command(
		fe.BinPath("xbusd"),
		"--config", xbusdCfgFilename,
		"--cert-dir", path.Join(fe.dirname, "certs", "xbusd"),
		"serve")

	xbusd.Dir = fe.dirname

	if err := captureCmdOutputs("xbusd", xbusd, fe.log); err != nil {
		return err
	}

	if err := xbusd.Start(); err != nil {
		return err
	}

	fe.xbusd = xbusd
	fe.log.Noticef("xbusd started")
	time.Sleep(250 * time.Millisecond)
	return nil
}

// StopXbusd stops the xbusd server
func (fe *FullEnv) StopXbusd() error {
	if fe.xbusd != nil {
		fe.xbusd.Process.Signal(os.Interrupt)
		var timer *time.Timer
		timer = time.AfterFunc(3*time.Second, func() {
			fe.log.Errorf("Xbusd still running (deadlocked ?). Attempting a SIGQUIT")
			if err := fe.xbusd.Process.Signal(syscall.SIGQUIT); err != nil {
				fe.log.Errorf("SIGQUIT failed with '%s', killing the process.", err)
				fe.xbusd.Process.Kill()
			} else {
				timer = time.AfterFunc(time.Second, func() {
					fe.log.Errorf("Xbusd still running. Killing it")
					fe.xbusd.Process.Kill()
				})
			}
		})

		defer timer.Stop()
		if err := fe.xbusd.Wait(); err != nil {
			return err
		}
		fe.xbusd = nil
	}
	return nil
}

// Close stops everything, including the server
func (fe *FullEnv) Close() error {
	fe.Shutdown()
	fe.StopXbusd()
	if !fe.options.NoDBLock {
		if err := fe.dblock.Unlock(); err != nil {
			fe.log.Errorf("Error releasing lock: %s", err)
		}
	}
	for _, name := range fe.actorServices {
		xbus.UnregisterActorService(name)
	}
	fe.actorServices = nil
	if fe.t != nil && fe.t.Failed() {
		fe.t.Logf("Test failed. All the files can be found in %s", fe.dirname)
	}
	return nil
}

// CmdRunner can run a command with args
type CmdRunner struct {
	Dir     string
	Command string
	Args    []string
	Log     xbus.Logger
}

// NewCmdRunner initializes a CmdRunner
func NewCmdRunner(log xbus.Logger, dir, command string, args ...string) CmdRunner {
	return CmdRunner{dir, command, args, log}
}

// Run runs the command with args
func (r CmdRunner) Run(args ...string) error {
	cmd := exec.Command(r.Command, append(r.Args[:], args...)...)
	cmd.Dir = r.Dir

	b, err := cmd.CombinedOutput()
	if err != nil {
		r.Log.Errorf("'%s': %s", strings.Join(args, " "), b)
		return fmt.Errorf("Could not run '%s': %s", strings.Join(args, " "), err)
	}

	r.Log.Noticef("%s: %s", strings.Join(args, " "), b)

	return nil
}

// RunReadLastLine runs the command with args and returns the last line of
// the command combined output
func (r CmdRunner) RunReadLastLine(args ...string) (string, error) {
	cmd := exec.Command(r.Command, append(r.Args[:], args...)...)
	cmd.Dir = r.Dir

	b, err := cmd.CombinedOutput()
	if err != nil {
		r.Log.Errorf("'%s': %s", strings.Join(args, " "), b)
		return "", fmt.Errorf("Could not run '%s': %s", strings.Join(args, " "), err)
	}

	lines := bytes.Split(b, []byte("\n"))
	var lastLine string
	for i := len(lines) - 1; i >= 0; i-- {
		if len(lines[i]) != 0 {
			lastLine = string(lines[i])
			break
		}
	}
	return lastLine, nil
}

// RunReadAndFindSubmatch runs the command with args, then match each line with the
// given regex and return the submatches for each
func (r CmdRunner) RunReadAndFindSubmatch(re *regexp.Regexp, args ...string) ([][]string, error) {
	cmd := exec.Command(r.Command, append(r.Args[:], args...)...)
	cmd.Dir = r.Dir

	b, err := cmd.CombinedOutput()
	if err != nil {
		r.Log.Errorf("'%s': %s", strings.Join(args, " "), b)
		return nil, fmt.Errorf("Could not run '%s': %s", strings.Join(args, " "), err)
	}

	var result [][]string

	lines := bytes.Split(b, []byte("\n"))

	for i := len(lines) - 1; i >= 0; i-- {
		subs := re.FindSubmatch(lines[i])
		if subs != nil {
			var groups []string
			for _, b := range subs[1:] {
				groups = append(groups, string(b))
			}
			result = append(result, groups)
		}
	}

	return result, nil
}

// MustRun runs the command with args and panic if it fails
func (r CmdRunner) MustRun(args ...string) {
	if err := r.Run(args...); err != nil {
		panic(err)
	}
}

func must(err error, msg ...string) {
	if err != nil {
		panic(fmt.Errorf("%s: %s", strings.Join(msg, " "), err))
	}
}

func catcher(err *error) {
	if r := recover(); r != nil {
		if e, ok := r.(error); ok {
			*err = e
		} else {
			panic(r)
		}
	}
}
