package process

import (
	"os"
	"testing"

	"gopkg.in/yaml.v2"

	"github.com/sergi/go-diff/diffmatchpatch"

	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/testutils"
)

type TestStep struct {
	Label       string
	LogEntries  []storage.ProcessLogEntry
	TargetState storage.ProcessState
}

type Test struct {
	Name         string
	Graph        string
	InitialState storage.ProcessState
	Steps        []TestStep
}

type TestFile struct {
	Graphs map[string]storage.Graph
	Tests  []Test
}

func LogYaml(t *testing.T, i interface{}) {
	b, err := yaml.Marshal(i)

	if err != nil {
		t.Error(err)
	}

	t.Logf(string(b))
}

func assertSameYaml(t *testing.T, fatal bool, expected, actual interface{}) {
	sExpected, err := yaml.Marshal(expected)
	if err != nil {
		t.Fatal(err)
	}

	sActual, err := yaml.Marshal(actual)
	if err != nil {
		t.Fatal(err)
	}

	var notEqual = false

	dmp := diffmatchpatch.New()

	diffs := dmp.DiffMain(string(sActual), string(sExpected), false)
	for _, diff := range diffs {
		if diff.Type != diffmatchpatch.DiffEqual {
			notEqual = true
		}
	}

	if notEqual {
		f := t.Errorf
		if fatal {
			f = t.Fatalf
		}
		f("Different yaml contents. Diff:\n%s", dmp.DiffPrettyText(diffs))
	}
}

func TestStateEngine(t *testing.T) {
	for _, filename := range []string{"state_engine_test_basics.yaml"} {
		t.Run(filename, func(t *testing.T) {
			t.Parallel()
			in, err := os.Open(filename)
			if err != nil {
				t.Fatal(err)
			}
			decoder := yaml.NewDecoder(in)
			var testFile TestFile
			if err := decoder.Decode(&testFile); err != nil {
				t.Fatal(err)
			}
			for _, test := range testFile.Tests {
				t.Run(test.Name, func(t *testing.T) {
					t.Parallel()

					graphName := test.Graph
					if graphName == "" {
						graphName = "default"
					}
					graphDef, ok := testFile.Graphs[graphName]
					if !ok {
						t.Fatalf(
							"No graph '%s' defined in the test file", graphName)
					}
					g := graph.NewGraph(graphDef)
					logger := testutils.GetTestLogger(t)
					state := test.InitialState

					for _, step := range test.Steps {
						t.Logf("Step %s", step.Label)

						for _, entry := range step.LogEntries {
							state, err = NextState(g, state, entry, logger)
							if err != nil {
								t.Fatal(err)
							}
						}

						assertSameYaml(t, true, step.TargetState, state)
					}

				})
			}
		})
	}
}
