// Copyright 2016 Orus.io. All rights reserved.

package processstate

import (
	"context"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	// ServiceName name of the service
	ServiceName = "processstate"

	// QueueName queue name for queuesubscribes
	QueueName = "xbus.services." + ServiceName
)

// Options is used to create a 'processstate' service
type Options struct {
	services.Options

	LogStorage     storage.LogStorage
	ProcessManager *process.Manager
}

// NewService creates a Service
func (o Options) NewService() *Service {
	return &Service{
		BaseService: services.NewBaseService(ServiceName, o.Options),
		o:           o,
	}
}

// Service is in charge of emitting process status events to various subjects
type Service struct {
	services.BaseService

	o Options

	api          *api.Client
	processState *api.ProcessStateHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	s.processState = api.NewProcessStateHandler(
		context.Background(), s.Conn, nil,
	)
	s.api = api.NewClient(s.Conn)
	sub, err := s.api.CoreEvents.ProcessSubscribe(s.handleCoreEventsProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)
	return nil
}

func (s *Service) handleCoreEventsProcess(evt api.CoreEvent) {
	s.TaskAdd()
	defer s.TaskDone()
	actor := api.Actor{ID: evt.Process.TriggerEmitterID}
	statusList := s.o.ProcessManager.GetEmitterEnvelopeStatus(
		actor, evt.Process.TriggerEnvelopeID,
	)
	if len(statusList) == 0 {
		panic("No process for trigger... which is weird if not impossible. Please report")
	}
	if len(statusList) != 1 {
		panic("Not implemented")
	}
	var envelopeState api.EmitterEnvelopeState
	for id, status := range statusList {
		process := api.EmitterEnvelopeState_Process{
			ID:     id,
			Status: status,
		}
		for _, entry := range s.o.LogStorage.GetProcessMessages(
			id, api.LogLevel_ERROR,
		) {
			s.Log.Noticef("%+v", entry)
			process.Errors = append(process.Errors, entry.Message)
		}
		envelopeState.Status = status
		envelopeState.ProcessList = append(envelopeState.ProcessList, process)
	}
	if err := s.processState.EnvelopePublish(
		evt.Process.TriggerEmitterID.String(),
		evt.Process.TriggerEnvelopeID.String(),
		envelopeState,
	); err != nil {
		s.Log.Errorf("Error sending process state: %s", err)
	}
}
