// Copyright 2016 Orus.io. All rights reserved.

package storage

import (
	"bitbucket.org/orus-io/go-xbus/api"
)

// EnvelopeFilter filters a batch of envelope ids at once
type EnvelopeFilter func([]api.UUID) []api.UUID

// EnvelopeStorage can store partial envelopes, track the completions of the events
// inside it, serve back the data in chunks
type EnvelopeStorage interface {
	// Store part or all fo an envelope
	// The returned EnvelopeStatus reflect the final state of the complete
	// envelope, once the given part is merge to it.
	//
	// Concurrency & ordering:
	// The envelope parts can be passed in any order, from different
	// goroutines, and to different instance of the Storage in different
	// processes all pointed to the same backend.
	//
	// Error handling:
	// It any error occur, including validity errors (which are reported
	// within the EnvelopeStatus), the envelope should be marked as 'error'
	// in the storage, and no other part of the envelope should be accepted.
	StoreEnvelope(api.Envelope) (EnvelopeStatus, error)

	// GetEventTypes returns the eventtypes of an envelope, or nil if unknown
	// exists is false if the envelope is unknown, true if exists.
	// should panic on any internal error.
	GetEventTypes(id api.UUID) (types []string, exists bool)

	ReadEnvelope(id api.UUID, position EnvelopeReadPos, maxsize int) (api.Envelope, EnvelopeReadPos, error)

	Purge(EnvelopeFilter)
}

// EventStatus contains the detailled status of an event
type EventStatus struct {
	ID             api.UUID
	Type           string
	AvailableItems uint64 // number of items that can be read in a sequence
	TotalItems     uint64 // total number of items or 0 if unknown
	Status         string // one of 'receiving', 'error', 'complete'
	Err            error
}

// Error implements the error interface in case Status is "error"
func (s EventStatus) Error() string {
	if s.Status != "error" {
		return s.Status
	}
	msg := "Event " + s.ID.String() + " error"
	if s.Err != nil {
		msg += s.Err.Error()
	}
	return msg
}

// EnvelopeStatus contains the detailled status of an envelope,
// allowing the broker to make decisions
type EnvelopeStatus struct {
	ID         api.UUID
	TypesKnown bool   // true if the type of all the events is known
	Status     string // one of 'receiving', 'error', 'complete'
	Err        error

	Events []EventStatus
}

// Error implements the error interface in case Status is "error"
func (s EnvelopeStatus) Error() string {
	if s.Status != "error" {
		return s.Status
	}
	msg := "Envelope " + s.ID.String() + " error: "
	if s.Err != nil {
		msg += s.Err.Error()
	}
	for _, evStatus := range s.Events {
		if evStatus.Status == "error" {
			msg += "\n  " + evStatus.Error()
		}
	}
	return msg
}

// EventTypes returns the event types or nil if not fully known yet
func (s EnvelopeStatus) EventTypes() []string {
	if !s.TypesKnown {
		return nil
	}
	var types []string
	for _, eventStatus := range s.Events {
		types = append(types, eventStatus.Type)
	}
	return types
}

// EventReadPos containts the current reading position of an event
type EventReadPos struct {
	Index     uint64
	ItemCount uint64
}

// EnvelopeReadPos contains the current reading position of each event of an envelope
// It can be used to retrieve the next part of an envelope later
type EnvelopeReadPos struct {
	Events   map[api.UUID]EventReadPos
	Start    bool
	Complete bool
}

// EnvelopeStartPos returns a new EnvelopeReadPos pointing the beginning of the
// envelope
func EnvelopeStartPos() EnvelopeReadPos {
	return EnvelopeReadPos{
		Events:   make(map[api.UUID]EventReadPos),
		Start:    true,
		Complete: false,
	}
}

// Clone duplicate the position
func (p EnvelopeReadPos) Clone() EnvelopeReadPos {
	newMap := make(map[api.UUID]EventReadPos)
	for id, ep := range p.Events {
		newMap[id] = ep
	}
	p.Events = newMap
	return p
}

// Equals returns true if postions are equals
func (p EnvelopeReadPos) Equals(other EnvelopeReadPos) bool {
	if p.Start != other.Start || p.Complete != other.Complete || len(p.Events) != len(other.Events) {
		return false
	}
	for id, pos := range p.Events {
		if other.Events[id] != pos {
			return false
		}
	}
	return true
}

// DecodePosition translate a api.EnvelopePosition into a storage.EnvelopeReadPos
func DecodePosition(apiPos api.EnvelopePosition) EnvelopeReadPos {
	pos := EnvelopeReadPos{
		Start:    apiPos.Start,
		Complete: apiPos.Complete,
		Events:   make(map[api.UUID]EventReadPos),
	}
	for _, event := range apiPos.EventPositions {
		pos.Events[event.EventID] = EventReadPos{
			Index:     event.Index,
			ItemCount: event.ItemCount,
		}
	}
	return pos
}

// EncodePosition translate a storage.EnvelopeReadPos into a api.EnvelopePosition
func EncodePosition(id api.UUID, pos EnvelopeReadPos) *api.EnvelopePosition {
	apiPos := api.EnvelopePosition{
		EnvelopeID: id,
		Start:      pos.Start,
		Complete:   pos.Complete,
	}
	for id, event := range pos.Events {
		apiPos.EventPositions = append(apiPos.EventPositions, &api.EventPosition{
			EventID:   id,
			Index:     event.Index,
			ItemCount: event.ItemCount,
		})
	}
	return &apiPos
}
