// Copyright 2016 Orus.io. All rights reserved.

package xbus

import (
	"bytes"
	"crypto/ecdsa"
	"crypto/elliptic"
	"crypto/rand"
	"crypto/x509/pkix"
	"encoding/pem"
	"errors"
	"fmt"

	api "bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

// ErrAlreadyExists is returned by CreateKey or CreateCSR when the key
// or cert already exists and force is false
var ErrAlreadyExists = errors.New("Key or Certificate already exists")

// CreateKey is used to create a new pub/priv key for the client.
// this will update the client options directly and only return the error code
func (c *Client) CreateKey(force bool) error {
	if c.options.PrivKey != nil && !force {
		return ErrAlreadyExists
	}
	curve := elliptic.P521()

	key, err := ecdsa.GenerateKey(curve, rand.Reader)
	if err != nil {
		return err
	}
	c.options.PrivKey = key
	c.options.PubKey = c.options.PrivKey.Public()
	c.options.ChangeCallback(c.options)
	return nil
}

// CreateCSR is used to create a certificate request, it must be used ONLY
// if you have a valid key pair in the client options
// host, org and cn are string that will be present in your certificate request
//
//	host: represent the hostname from which we expect
//	this certificate to be used
//	this can eventually be an IP address if you do not have a hostname
//
//	org is just used as an indication
//
//	cn should be the "name" of your client
//
func (c *Client) CreateCSR(host, org, cn string, overwrite bool) (err error) {
	if c.options.CSR != nil && !overwrite {
		return ErrAlreadyExists
	}
	certSubject := pkix.Name{
		Organization: []string{org},
		CommonName:   cn,
	}
	derBytes, err := tools.CreateCSR(c.options.PrivKey, host, certSubject)
	if err != nil {
		return err
	}
	c.options.CSR = derBytes
	c.options.Certificate = nil
	c.options.ChangeCallback(c.options)
	return nil
}

// Register does a registration request to the bus
func (c *Client) Register() (api.Registration_Status, error) {
	var (
		req api.RegistrationRequest
		reg api.Registration
	)
	if !c.authenticated {
		b := bytes.NewBufferString("")
		err := pem.Encode(
			b,
			&pem.Block{
				Type:  "CERTIFICATE REQUEST",
				Bytes: c.options.CSR,
			},
		)

		if err != nil {
			return api.Registration_PENDING, err
		}

		req.Name = c.options.Name
		req.Type = c.options.Type
		req.CSR = b.String()
	}

	req.ActorList = make([]api.Actor, 0, len(c.options.ActorList))
	for _, actorOptions := range c.options.ActorList {
		req.ActorList = append(req.ActorList, actorOptions.Actor)
	}

	var err error
	if c.authenticated {
		reg, err = c.API.SvcRegistration.Update(c.ID().String(), req)
	} else {
		reg, err = c.API.SvcRegistration.Register(req)
	}

	if err != nil {
		return api.Registration_PENDING, err
	}

	if c.options.ID.IsZero() && !reg.ID.IsZero() {
		c.options.ID = reg.ID
	}
	if c.options.Certificate == nil && reg.SignedCertificate != "" {
		b, _ := pem.Decode([]byte(reg.SignedCertificate))
		if b == nil {
			return 0, fmt.Errorf("Could not decode PEM SignedCertificate")
		}
		c.options.Certificate = b.Bytes
	}
	if reg.ServerCA != "" {
		b, _ := pem.Decode([]byte(reg.ServerCA))
		if b == nil {
			return 0, fmt.Errorf("Could not decode PEM ServerCA")
		}
		c.options.RootCAs = [][]byte{b.Bytes}
	}

	actorOptionsByName := make(map[string]ActorOptions)
	for _, actorOptions := range c.options.ActorList {
		actorOptionsByName[actorOptions.Name] = actorOptions
	}
	c.options.ActorList = make([]ActorOptions, 0, len(reg.ActorList))
	for _, actor := range reg.ActorList {
		opts, _ := actorOptionsByName[actor.Name]
		opts.Actor = actor
		c.options.ActorList = append(c.options.ActorList, opts)
	}

	// Client options have changed; signal that (when a callback is set).
	if c.options.ChangeCallback != nil {
		c.options.ChangeCallback(c.options)
	}

	return reg.RegistrationStatus, nil
}
